@@ -2,6 +2,7 @@ import cgi
 import functools
 import urllib
 
+from django.core.exceptions import ImproperlyConfigured
 from django.conf import settings
 from django.core.cache import cache
 from django.utils.encoding import iri_to_uri
@@ -13,6 +14,10 @@ from django.utils.cache import (
 )
 
 
+REMEMBERED_URLS_KEY = 'fancy-urls'
+LONG_TIME = 60 * 60 * 24 * 30
+
+
 class RequestPath(object):
     def __init__(self, request, only_get_keys):
         self.request = request
@@ -95,7 +100,17 @@ class UpdateCacheMiddleware(object):
                 )
 
             with RequestPath(request, self.only_get_keys):
-                cache_key = learn_cache_key(request, response, timeout, key_prefix)
+                cache_key = learn_cache_key(
+                    request,
+                    response,
+                    timeout,
+                    key_prefix
+                )
+
+                if self.remember_all_urls:
+                    self.remember_url(request, cache_key, timeout)
+
+            #print "SET %s (%s)" %(request.path, timeout)
             cache.set(cache_key, response, timeout)
 
         if self.post_process_response_always:
@@ -106,6 +121,16 @@ class UpdateCacheMiddleware(object):
 
         return response
 
+    def remember_url(self, request, cache_key, timeout):
+        url = request.get_full_path()
+        remembered_urls = cache.get(REMEMBERED_URLS_KEY, {})
+        remembered_urls[url] = cache_key
+        cache.set(
+            REMEMBERED_URLS_KEY,
+            remembered_urls,
+            LONG_TIME
+        )
+
 
 class FetchFromCacheMiddleware(object):
     """
@@ -120,15 +145,43 @@ class FetchFromCacheMiddleware(object):
         Checks whether the page is already cached and returns the cached
         version if available.
         """
+        response = self._process_request(request)
+        if self.remember_stats_all_urls:
+            # then we're nosy
+            cache_key = request.get_full_path()
+            if response is None:
+                cache_key += '__misses'
+                #print "**Cache Miss**", request.path
+            else:
+                cache_key += '__hits'
+            if cache.get(cache_key) is None:
+                cache.set(cache_key, 0, LONG_TIME)
+            cache.incr(cache_key)
+                #print "**Cache HIT!**", request.path
+        return response
+
+    def _process_request(self, request):
         if self.cache_anonymous_only:
-            assert hasattr(request, 'user'), "The Django cache middleware with CACHE_MIDDLEWARE_ANONYMOUS_ONLY=True requires authentication middleware to be installed. Edit your MIDDLEWARE_CLASSES setting to insert 'django.contrib.auth.middleware.AuthenticationMiddleware' before the CacheMiddleware."
+            if not hasattr(request, 'user'):
+                raise ImproperlyConfigured(
+                    "The Django cache middleware with "
+                    "CACHE_MIDDLEWARE_ANONYMOUS_ONLY=True requires "
+                    "authentication middleware to be installed. Edit your "
+                    "MIDDLEWARE_CLASSES setting to insert "
+                    "'django.contrib.auth.middleware.AuthenticationMiddleware'"
+                    "before the CacheMiddleware."
+                )
 
         if not request.method in ('GET', 'HEAD'):
             request._cache_update_cache = False
             # Don't bother checking the cache.
             return None
 
-        if request.GET and not callable(self.key_prefix) and not self.only_get_keys:
+        if (
+            request.GET and
+            not callable(self.key_prefix) and
+            not self.only_get_keys
+        ):
             request._cache_update_cache = False
             # Default behaviour for requests with GET parameters: don't bother
             # checking the cache.
@@ -165,7 +218,10 @@ class FetchFromCacheMiddleware(object):
 
         request._cache_update_cache = False
         if self.post_process_response_always:
-            response = self.post_process_response_always(response, request=request)
+            response = self.post_process_response_always(
+                response,
+                request=request
+            )
 
         return response
 
@@ -197,8 +253,8 @@ class CacheMiddleware(UpdateCacheMiddleware, FetchFromCacheMiddleware):
         before the response gets set in cache.
 
     :param post_process_response_always:
-        Callable function that gets called with the response (and request) every
-        time the response goes through the middleware cached or not.
+        Callable function that gets called with the response (and request)
+        every time the response goes through the middleware cached or not.
 
     :param only_get_keys:
         List of query string keys to reduce the cache key to. Without this a
@@ -206,6 +262,14 @@ class CacheMiddleware(UpdateCacheMiddleware, FetchFromCacheMiddleware):
         different cache keys when it could be that the `other=junk` parameter
         doesn't change anything.
 
+    :param remember_all_urls:
+        With this option you can have all cached URLs stored in cache which
+        can make it easy to do things like cache invalidation by URL.
+
+    :param remember_stats_all_urls:
+        Only applicable if `remember_all_urls` is set. This stores a count
+        of the number of times a `cache_page` hits and misses.
+
     """
     def __init__(self,
                  cache_timeout=settings.CACHE_MIDDLEWARE_SECONDS,
@@ -218,7 +282,18 @@ class CacheMiddleware(UpdateCacheMiddleware, FetchFromCacheMiddleware):
                  patch_headers=False,
                  post_process_response=None,
                  post_process_response_always=None,
-                 only_get_keys=None):
+                 only_get_keys=None,
+                 remember_all_urls=getattr(
+                     settings,
+                     'FANCY_REMEMBER_ALL_URLS',
+                     False
+                 ),
+                 remember_stats_all_urls=getattr(
+                     settings,
+                     'FANCY_REMEMBER_STATS_ALL_URLS',
+                     False
+                 ),
+                 ):
         self.patch_headers = patch_headers
         self.cache_timeout = cache_timeout
         self.key_prefix = key_prefix
@@ -228,3 +303,5 @@ class CacheMiddleware(UpdateCacheMiddleware, FetchFromCacheMiddleware):
         if isinstance(only_get_keys, basestring):
             only_get_keys = [only_get_keys]
         self.only_get_keys = only_get_keys
+        self.remember_all_urls = remember_all_urls
+        self.remember_stats_all_urls = remember_stats_all_urls
